import osmnx as ox
import networkx as nx
import itertools
import pickle
import time, datetime
import os.path


ox.settings.log_console=True
ox.settings.use_cache=True

print('OSMnx version: ', ox.__version__)
print('NetworkX version: ',nx.__version__)

start_time = time.time()

def elapsed_time(start):
    end_time = time.time()
    duration = round(end_time - start)
    timestamp = str(datetime.timedelta(seconds = duration))
    print(timestamp)



def graph_length(Q):
    length = ox.stats.edge_length_total(Q) #in meters
    length = length * 0.0006213712
    return length
       
def report(Q, txt=''):
    print(len(Q.nodes()), ' nodes ', txt)
    print(len(Q.edges()), ' edges ', txt)
    if isinstance(Q,nx.MultiGraph):    
        length = ox.stats.edge_length_total(Q) #in meters
        length = length * 0.0006213712
        print(f'total length {length:.2f} miles ', txt)
    zeroes = 0    
    odds = 0
    evens = 0
    twos = 0
    ones = 0
    for node in Q.nodes:
        d = Q.degree(node)
        if d == 0:
            zeroes += 1
        if d % 2 == 1:
            odds += 1
        else:
            evens += 1
        if d == 2:
            twos += 1
        if d == 1:
            ones += 1
    if zeroes > 0:            
        print('number of zero-degree nodes ' + txt + ' :', zeroes)            
    print('number of odd intersections ' + txt + ' :', odds)
    print('number of even intersections ' + txt + ' :', evens)
    print('number of 2-way intersections ' + txt + ' :', twos)
    print('number of trailhead/terminus ' + txt + ' :', ones)
    print('')
        


H = pickle.load(open('h_graph.pkl', 'rb'))
report(H,'graph loaded from disk')

# we expect some disconnected subgraphs, so choose the largest
nodes = max(nx.connected_components(H), key=len)
J = nx.subgraph(H, nodes).copy()

# remove trailheads and termini
def delete_ends(Q):
    while True: # simulate do-while
        rn = []
        for node, neighbors in Q.adjacency():
            if len(neighbors) == 1:
                rn.append(node)
        if len(rn)>0:            
            Q.remove_nodes_from(rn)
        else:
            break

delete_ends(J)

T = J.to_directed()
T = ox.simplification.simplify_graph(T, strict=True, remove_rings=True, track_merged=True)
J = T.to_undirected()
            
        
# get a graph of just the odd nodes, improve later to handle 4-way intersections

def odd_nodes_graph(Q):
    R = Q.copy()
    unwanted_edges = []
    for u, v, d in R.edges(data=True):
        if R.degree(u) % 2 != 1 or R.degree(v) % 2 != 1:
            unwanted_edges.append( (u,v) )
    R.remove_edges_from( unwanted_edges )
    return R
    

K = odd_nodes_graph(J)           
#report(K,'in odd-nodes graph')
#draw(K,'Odd-nodes Graph')


        


# add in edges that bypass 4-nodes in graph
# fix: 4-way connected to a 4-way
# fix: 4-way connected to a 4-way
def complete_3node_graph(Q,R):
    n4list = []
    for node, neighbors in Q.adjacency():
        if len(neighbors) == 4:
            n4list.append( (node,neighbors) )
            nlist = []
            llen = []
            for neighbor, edict in neighbors.items():
                w = min([d['length'] for d in edict.values()])
                nlist.append(neighbor)
                llen.append(w)
              #  for dic in edict.items():
                    #w = dic['length']
                    #nlist.append(neighbor)
                    #llen.append(w)
            # could use itertools?
            for i,j in [ (0,1), (1,2), (2,3), (3,0), (0,2),( 1,3) ]:
                R.add_edge(nlist[i],nlist[j], length= llen[i]+llen[j]) 
    return n4list

fix_4_list = complete_3node_graph(J,K)

report(K,'in odd-nodes 4node bypass')

def restore_4node_graph(Q,fix_4_list):
    for node, neighbors in fix_4_list:
       i = 0
       nlist = []
       vlist = []
       for neighbor, edict in neighbors.items():
           # for dic in edict:
           nlist.append(neighbor)
           vlist.append(edict[0])
       for i,j in [ (0,1), (1,2), (2,3), (3,0), (0,2),( 1,3) ]:               
            if Q.has_edge(nlist[i],nlist[j]):
                Q.remove_edge(nlist[i],nlist[j])
                if not Q.has_edge(node,nlist[i]):
                    Q.add_edge(node,nlist[i],**vlist[i]) #fix
                if not Q.has_edge(node,nlist[j]):                                  
                    Q.add_edge(node,nlist[j],**vlist[j])



max_length = 0
all_eulerian_matchings = []


# https://stackoverflow.com/questions/51933046/python-networkx-get-unique-matching-combinations#51934312
# This function is a thing of beauty, way beyond my abilities.
# I tried building my own, with much wailing, before finding this concise gem.
def all_maximal_matchings(T):

    maximal_matchings = []
    partial_matchings = [{(u,v)} for (u,v) in T.edges()]

    while partial_matchings:
        # get current partial matching
        m = partial_matchings.pop()
        nodes_m = set(itertools.chain(*m))

        extended = False
        for (u,v) in T.edges():
            if u not in nodes_m and v not in nodes_m:
                extended = True
                # copy m, extend it and add it to the list of partial matchings
                m_extended = set(m)
                m_extended.add((u,v))
                partial_matchings.append(m_extended)

        if not extended and m not in maximal_matchings:
            maximal_matchings.append(m)
            print('matching ', len(maximal_matchings))
            try_matching(m, K, fix_4_list, len(maximal_matchings))
            elapsed_time(start_time)

        if len(maximal_matchings) > 1000:  # fix
            print('Timeout on maximal matchings, ')
            break;

    return maximal_matchings

def try_matching(matching, Q, fix_4_list, index):
    T = Q.copy()
    T.remove_edges_from(matching)
    restore_4node_graph(T,fix_4_list)
    nodes = max(nx.connected_components(T), key=len)
    T2 = nx.subgraph(T, nodes).copy()
    delete_ends(T2) #just in case, probably not needed?
    if nx.is_eulerian(T2):
        all_eulerian_matchings.append(matching)
        print('found eulerian on maximal matching candidate ', index)
        print('length of successful matching: ', len(matching))
        report(T2,'matching ' + str(index))
        max_length = max(max_length,graph_length(T2))
    else:
        print('not eulerian matching, length ', len(matching))


all_matchings = all_maximal_matchings(K)
print('finished finding possible matchings')
pickle.dump(all_matchings,open('all_matchings.pkl', 'wb'))
pickle.dump(fix_4_list,open('fix_4_list.pkl', 'wb'))
pickle.dump(K,open('k_graph.pkl', 'wb'))

elapsed_time(start_time)

print('finished script ' + os.path.basename(__file__))






